
################################################################################
#                            skforecast.datsets                                #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

import pandas as pd
import textwrap


def fetch_dataset(
    name : str,
    version: str = 'latest',
    raw: bool = False,
    kwargs_read_csv: dict = {},
    verbose: bool = True
) -> pd.DataFrame:
    """
    Fetch a dataset from the skforecast-datasets repository. Available datasets
    are: 'h2o', 'items_sales', 'air_pollution', 'fuel_consumption', 'web_visits',
    'bike_sharing', 'store_item_demand'.

    Parameters
    ----------
    name: str
        Name of the dataset to fetch.
    version: str, int, default `'latest'`
        Version of the dataset to fetch. If 'latest', the lastest version will be 
        fetched (the one in the main branch). For a list of available versions, 
        see the repository branches.
    raw: bool, default `False`
        If True, the raw dataset is fetched. If False, the preprocessed dataset 
        is fetched. The preprocessing consists of setting the column with the 
        date/time as index and converting the index to datetime. A frequency is 
        also set to the index.
    kwargs_read_csv: dict, default `{}`
        Kwargs to pass to pandas `read_csv` function.
    verbose: bool, default `True`
        If True, print information about the dataset.
    
    Returns
    -------
    df: pandas DataFrame
        Dataset.
    
    """

    version = 'main' if version == 'latest' else f'{version}'

    datasets = {
        'h2o': {
            'url': (
                f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/'
                f'skforecast-datasets/{version}/data/h2o.csv'
            ),
            'sep': ',',
            'index_col': 'fecha',
            'date_format': '%Y-%m-%d',
            'freq': 'MS',
            'description': (
                'Monthly expenditure ($AUD) on corticosteroid drugs that the '
                'Australian health system had between 1991 and 2008. '
            ),
            'source': (
                'Hyndman R (2023). fpp3: Data for Forecasting: Principles and Practice'
                '(3rd Edition). http://pkg.robjhyndman.com/fpp3package/,'
                'https://github.com/robjhyndman/fpp3package, http://OTexts.com/fpp3.'
            )
        },
        'h2o_exog': {
            'url': (
                f"https://raw.githubusercontent.com/JoaquinAmatRodrigo/"
                f"skforecast-datasets/{version}/data/h2o_exog.csv"
            ),
            'sep': ',',
            'index_col': 'fecha',
            'date_format': '%Y-%m-%d',
            'freq': 'MS',
            'description': (
                'Monthly expenditure ($AUD) on corticosteroid drugs that the '
                'Australian health system had between 1991 and 2008. Two additional '
                'variables (exog_1, exog_2) are simulated.'
            ),
            'source': (
                "Hyndman R (2023). fpp3: Data for Forecasting: Principles and Practice "
                "(3rd Edition). http://pkg.robjhyndman.com/fpp3package/, "
                "https://github.com/robjhyndman/fpp3package, http://OTexts.com/fpp3."
            )
        },
        'fuel_consumption': {
            'url' : (
                f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/'
                f'skforecast-datasets/{version}/data/consumos-combustibles-mensual.csv'
            ),
            'sep': ',',
            'index_col': 'Fecha',
            'date_format': '%Y-%m-%d',
            'freq': 'MS',
            'description': (
                'Monthly fuel consumption in Spain from 1969-01-01 to 2022-08-01.'
            ),
            'source': (
                'Obtained from Corporación de Reservas Estratégicas de Productos '
                'Petrolíferos and Corporación de Derecho Público tutelada por el '
                'Ministerio para la Transición Ecológica y el Reto Demográfico. '
                'https://www.cores.es/es/estadisticas'
            )
        },
        'items_sales': {
            'url': (
                f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/'
                f'skforecast-datasets/{version}/data/simulated_items_sales.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d',
            'freq': 'D',
            'description': 'Simulated time series for the sales of 3 different items.',
            'source': 'Simulated data.'
        },
        'air_quality_valencia': {
            'url': (
                f"https://raw.githubusercontent.com/JoaquinAmatRodrigo/"
                f"skforecast-datasets/{version}/data/air_quality_valencia.csv"
            ),
            'sep': ',',
            'index_col': 'datetime',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'H',
            'description': (
                'Hourly measures of several air chemical pollutant (pm2.5, co, no, '
                'no2, pm10, nox, o3, so2) at Valencia city.'
            ),
            'source': (
                " Red de Vigilancia y Control de la Contaminación Atmosférica, "
                "46250054-València - Centre, "
                "https://mediambient.gva.es/es/web/calidad-ambiental/datos-historicos."
            )
        },
        'website_visits': {
            'url' : (
                f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/'
                f'skforecast-datasets/{version}/data/visitas_por_dia_web_cienciadedatos.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d',
            'freq': '1D',
            'description': (
                'Daily visits to the cienciadedatos.net website registered with the '
                'google analytics service.'
            ),
            'source': (
                "Amat Rodrigo, J. (2021). cienciadedatos.net (1.0.0). Zenodo. "
                "https://doi.org/10.5281/zenodo.10006330"
            )
        },
        'bike_sharing': {
            'url' : (
                f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/'
                f'skforecast-datasets/{version}/data/bike_sharing_dataset_clean.csv'
            ),
            'sep': ',',
            'index_col': 'date_time',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'H',
            'description': (
                'Hourly usage of the bike share system in the city of Washington D.C. '
                'during the years 2011 and 2012. In addition to the number of users per '
                'hour, information about weather conditions and holidays is available.'
            ),
            'source': (
                "Fanaee-T,Hadi. (2013). Bike Sharing Dataset. UCI Machine Learning "
                "Repository. https://doi.org/10.24432/C5W894."
            )
        },
        'bike_sharing_extended_features': {
            'url' : (
                f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/'
                f'skforecast-datasets/{version}/data/bike_sharing_extended_features.csv'
            ),
            'sep': ',',
            'index_col': 'date_time',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'H',
            'description': (
                'Hourly usage of the bike share system in the city of Washington D.C. '
                'during the years 2011 and 2012. In addition to the number of users per '
                'hour, the dataset was enriched by introducing supplementary features. '
                'Addition includes calendar-based variables (day of the week, hour of '
                'the day, month, etc.), indicators for sunlight, incorporation of '
                'rolling temperature averages, and the creation of polynomial features '
                'generated from variable pairs. All cyclic variables are encoded using '
                'sine and cosine functions to ensure accurate representation.'
            ),
            'source': (
                "Fanaee-T,Hadi. (2013). Bike Sharing Dataset. UCI Machine Learning "
                "Repository. https://doi.org/10.24432/C5W894."
            )
        },
        'australia_tourism': {
            'url' : (
                f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/'
                f'skforecast-datasets/{version}/data/australia_tourism.csv'
            ),
            'sep': ',',
            'index_col': 'date_time',
            'date_format': '%Y-%m-%d',
            'freq': 'Q',
            'description': (
                "Quarterly overnight trips (in thousands) from 1998 Q1 to 2016 Q4 "
                "across Australia. The tourism regions are formed through the "
                "aggregation of Statistical Local Areas (SLAs) which are defined by "
                "the various State and Territory tourism authorities according to "
                "their research and marketing needs."
            ),
            'source': (
                "Wang, E, D Cook, and RJ Hyndman (2020). A new tidy data structure to "
                "support exploration and modeling of temporal data, Journal of "
                "Computational and Graphical Statistics, 29:3, 466-478, "
                "doi:10.1080/10618600.2019.1695624."
            )
        },
        'uk_daily_flights': {
            'url': (
                f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/'
                f'skforecast-datasets/{version}/data/uk_daily_flights.csv'
            ),
            'sep': ',',
            'index_col': 'Date',
            'date_format': '%d/%m/%Y',
            'freq': 'D',
            'description': 'Daily number of flights in UK from 02/01/2019 to 23/01/2022.',
            'source': (
                'Experimental statistics published as part of the Economic activity and '
                'social change in the UK, real-time indicators release, Published 27 '
                'January 2022. Daily flight numbers are available in the dashboard '
                'provided by the European Organisation for the Safety of Air Navigation '
                '(EUROCONTROL). '
                'https://www.ons.gov.uk/economy/economicoutputandproductivity/output/'
                'bulletins/economicactivityandsocialchangeintheukrealtimeindicators/latest'
            )
        },
        'wikipedia_visits': {
            'url': (
                f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/'
                f'skforecast-datasets/{version}/data/wikipedia_visits.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d',
            'freq': 'D',
            'description': (
                'Log daily page views for the Wikipedia page for Peyton Manning. '
                'Scraped data using the Wikipediatrend package in R.'
            ),
            'source': (
                'https://github.com/facebook/prophet/blob/main/examples/'
                'example_wp_log_peyton_manning.csv'
            )
        },
        'vic_electricity': {
            'url': (
                f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/'
                f'skforecast-datasets/{version}/data/vic_electricity.csv'
            ),
            'sep': ',',
            'index_col': 'Time',
            'date_format': '%Y-%m-%dT%H:%M:%SZ',
            'freq': '30min',
            'description': 'Half-hourly electricity demand for Victoria, Australia',
            'source': (
                "O'Hara-Wild M, Hyndman R, Wang E, Godahewa R (2022).tsibbledata: Diverse "
                "Datasets for 'tsibble'. https://tsibbledata.tidyverts.org/, "
                "https://github.com/tidyverts/tsibbledata/. "
                "https://tsibbledata.tidyverts.org/reference/vic_elec.html"
            )
        },
        'store_sales': {
            'url': (
                f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/'
                f'skforecast-datasets/{version}/data/store_sales.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d',
            'freq': '1D',
            'description': (
                'This dataset contains 913,000 sales transactions from 2013-01-01 to '
                '2017-12-31 for 50 products (SKU) in 10 stores.'
            ),
            'source': (
                'The original data was obtained from: inversion. (2018). Store Item '
                'Demand Forecasting Challenge. Kaggle. '
                'https://kaggle.com/competitions/demand-forecasting-kernels-only'
            )
        }
    }
    
    if name not in datasets.keys():
        raise ValueError(
            f"Dataset {name} not found. Available datasets are: {list(datasets.keys())}"
        )
    
    url = datasets[name]['url']

    try:
        sep = datasets[name]['sep']
        df = pd.read_csv(url, sep=sep, **kwargs_read_csv)
    except:
        raise ValueError(
            f"Error reading dataset {name} from {url}. Try to version = 'latest'"
        )

    if not raw:
        try:
            index_col = datasets[name]['index_col']
            freq = datasets[name]['freq']
            if freq == 'H' and pd.__version__ >= '2.2.0':
                freq = "h"
            date_format = datasets[name]['date_format']
            df = df.set_index(index_col)
            df.index = pd.to_datetime(df.index, format=date_format)
            df = df.asfreq(freq)
            df = df.sort_index()
        except:
            pass
    
    if verbose:
        print(name)
        print('-'*len(name))
        description = textwrap.fill(datasets[name]['description'], width=80)
        source = textwrap.fill(datasets[name]['source'], width=80)
        print(description)
        print(source)
        print(f"Shape of the dataset: {df.shape}")

    return df


def load_demo_dataset(version: str = 'latest') -> pd.Series:
    """
    Load demo data set with monthly expenditure ($AUD) on corticosteroid drugs that
    the Australian health system had between 1991 and 2008. Obtained from the book:
    Forecasting: Principles and Practice by Rob J Hyndman and George Athanasopoulos.
    Index is set to datetime with monthly frequency and sorted.

    Parameters
    ----------
    version: str, default `'latest'`
        Version of the dataset to fetch. If 'latest', the lastest version will be
        fetched (the one in the main branch). For a list of available versions,
        see the repository branches.

    Returns
    -------
    df: pandas Series
        Dataset.
    
    """

    version = 'main' if version == 'latest' else f'{version}'

    url = (
        f'https://raw.githubusercontent.com/JoaquinAmatRodrigo/skforecast-datasets/{version}/'
        'data/h2o.csv'
    )

    df = pd.read_csv(url, sep=',', header=0, names=['y', 'datetime'])
    df['datetime'] = pd.to_datetime(df['datetime'], format='%Y-%m-%d')
    df = df.set_index('datetime')
    df = df.asfreq('MS')
    df = df['y']
    df = df.sort_index()

    return df



# Unit test _create_train_X_y ForecasterAutoregMultiVariate
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.ForecasterAutoregMultiVariate import ForecasterAutoregMultiVariate
from skforecast.exceptions import MissingValuesWarning
from sklearn.linear_model import LinearRegression
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder


def test_create_train_X_y_TypeError_when_series_not_DataFrame():
    """
    Test TypeError is raised when series is not a pandas DataFrame.
    """
    series = pd.Series(np.arange(7))
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1', lags=3, steps=2)

    err_msg = re.escape(
        f"`series` must be a pandas DataFrame. Got {type(series)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster._create_train_X_y(series=series)


def test_create_train_X_y_ValueError_when_level_not_in_series():
    """
    Test ValueError is raised when level not in series column names.
    """
    series = pd.DataFrame({'1': pd.Series(np.arange(5)),  
                           'l2': pd.Series(np.arange(5))})
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1', lags=3, steps=2)

    series_col_names = list(series.columns)

    err_msg = re.escape(
        (f"One of the `series` columns must be named as the `level` of the forecaster.\n"
         f"  Forecaster `level` : {forecaster.level}.\n"
         f"  `series` columns   : {series_col_names}.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y(series=series)


def test_create_train_X_y_ValueError_when_lags_keys_not_in_series():
    """
    Test ValueError is raised when lags keys are not in series column names.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)),  
                           'l2': pd.Series(np.arange(10))})
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                               lags={'l1': 2, '2': 3}, steps=2)
    
    series_col_names = list(series.columns)

    err_msg = re.escape(
        (f"When `lags` parameter is a `dict`, its keys must be the "
         f"same as `series` column names.\n"
         f"  Lags keys        : {list(forecaster.lags_.keys())}.\n"
         f"  `series` columns : {series_col_names}.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y(series=series)


def test_create_train_X_y_ValueError_when_len_series_is_lower_than_maximum_lag_plus_steps():
    """
    Test ValueError is raised when length of y is lower than maximum lag plus
    number of steps included in the forecaster.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(5)),  
                           'l2': pd.Series(np.arange(5))})
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                               lags=3, steps=3)

    err_msg = re.escape(
        (f"Minimum length of `series` for training this forecaster is "
         f"{forecaster.max_lag + forecaster.steps}. Got {len(series)}. Reduce the "
         f"number of predicted steps, {forecaster.steps}, or the maximum "
         f"lag, {forecaster.max_lag}, if no more data is available.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y(series=series)


def test_create_train_X_y_IgnoredArgumentWarning_when_levels_of_transformer_series_not_equal_to_series_col_names():
    """
    Test IgnoredArgumentWarning is raised when `transformer_series` is a dict and its keys 
    are not the same as series column names.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)),  
                           'l2': pd.Series(np.arange(10))})
    dict_transformers = {'l1': StandardScaler(), 
                         'l3': StandardScaler()}
    
    forecaster = ForecasterAutoregMultiVariate(
                     regressor          = LinearRegression(), 
                     level              = 'l1',
                     lags               = 3,
                     steps              = 3,
                     transformer_series = dict_transformers
                 )
    series_not_in_transformer_series = set(series.columns) - set(forecaster.transformer_series.keys())

    warn_msg = re.escape(
        (f"{series_not_in_transformer_series} not present in `transformer_series`."
         f" No transformation is applied to these series.")
    )
    with pytest.warns(UserWarning, match = warn_msg):
        forecaster._create_train_X_y(series=series)


def test_create_train_X_y_TypeError_when_exog_is_categorical_of_no_int():
    """
    Test TypeError is raised when exog is categorical with no int values.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(6)),  
                           'l2': pd.Series(np.arange(6))})
    exog = pd.Series(['A', 'B', 'C', 'A', 'B', 'C'], name='exog', dtype='category')
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1', lags=2, steps=2)

    err_msg = re.escape(
        ("If exog is of type category, it must contain only integer values. "
         "See skforecast docs for more info about how to include categorical "
         "features https://skforecast.org/"
         "latest/user_guides/categorical-features.html")
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster._create_train_X_y(series=series, exog=exog)


def test_create_train_X_y_MissingValuesWarning_when_exog_has_missing_values():
    """
    Test _create_train_X_y is issues a MissingValuesWarning when exog has missing values.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(6)),  
                           'l2': pd.Series(np.arange(6))})
    exog = pd.Series([1, 2, 3, np.nan, 5, 6], name='exog')
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l2', lags=2, steps=2)

    warn_msg = re.escape(
        ("`exog` has missing values. Most machine learning models do "
         "not allow missing values. Fitting the forecaster may fail.")  
    )
    with pytest.warns(MissingValuesWarning, match = warn_msg):
        forecaster._create_train_X_y(series=series, exog=exog)


def test_create_train_X_y_TypeError_when_exog_is_not_pandas_Series_or_DataFrame():
    """
    Test TypeError is raised when exog is not a pandas Series or DataFrame.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(10))})
    exog = np.arange(10)
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1', 
                                               lags=3, steps=2)
    
    err_msg = (
        f"`exog` must be a pandas Series or DataFrame. Got {type(exog)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster._create_train_X_y(series=series, exog=exog)


@pytest.mark.parametrize('exog', ['l1', ['l1'], ['l1', 'l2']])
def test_create_train_X_y_ValueError_when_exog_columns_same_as_series_col_names(exog):
    """
    Test ValueError is raised when an exog column is named the same as
    the series columns.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(10))})

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1', lags=3, steps=2)
    series_col_names = list(series.columns)
    exog_col_names = exog if isinstance(exog, list) else [exog]

    err_msg = re.escape(
        (f"`exog` cannot contain a column named the same as one of "
         f"the series (column names of series).\n"
         f"  `series` columns : {series_col_names}.\n"
         f"  `exog`   columns : {exog_col_names}.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y(series=series, exog=series[exog])


@pytest.mark.parametrize("exog", 
                         [pd.Series(np.arange(15)), 
                          pd.DataFrame(np.arange(50).reshape(25, 2))])
def test_create_train_X_y_ValueError_when_series_and_exog_have_different_length(exog):
    """
    Test ValueError is raised when length of series is not equal to length exog.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(10))})
    forecaster = forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                                            lags=3, steps=3)

    err_msg = re.escape(
        (f"`exog` must have same number of samples as `series`. "
         f"length `exog`: ({len(exog)}), length `series`: ({len(series)})")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y(series=series, exog=exog)


def test_create_train_X_y_ValueError_when_series_and_exog_have_different_index():
    """
    Test ValueError is raised when series and exog have different index.
    """
    series = pd.DataFrame(
        {"l1": pd.Series(np.arange(10)), "l2": pd.Series(np.arange(10))}
    )
    series.index = pd.date_range(start="2022-01-01", periods=10, freq="1D")
    forecaster = ForecasterAutoregMultiVariate(
        LinearRegression(), level="l1", lags=3, steps=3
    )

    exog = pd.Series(
        np.arange(10), index=pd.RangeIndex(start=0, stop=10, step=1), name="exog"
    )

    err_msg = re.escape(
        ("Different index for `series` and `exog`. They must be equal "
         "to ensure the correct alignment of values.")
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster.fit(series=series, exog=exog)


@pytest.mark.parametrize("level, expected_y_values", 
                         [('l1', [-0.52223297, -0.17407766,  0.17407766,  0.52223297,  0.87038828, 1.21854359,  1.5666989 ]), 
                          ('l2', [-0.52223297, -0.17407766,  0.17407766,  0.52223297,  0.87038828, 1.21854359,  1.5666989 ])])
def test_create_train_X_y_output_when_lags_3_steps_1_and_exog_is_None(level, expected_y_values):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 3 and steps is 1.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(100, 110), dtype=float)})
    exog = None

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level=level,
                                               lags=3, steps=1)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[-0.87038828, -1.21854359, -1.5666989 , -0.87038828, -1.21854359, -1.5666989 ],
                             [-0.52223297, -0.87038828, -1.21854359, -0.52223297, -0.87038828, -1.21854359],
                             [-0.17407766, -0.52223297, -0.87038828, -0.17407766, -0.52223297, -0.87038828],
                             [ 0.17407766, -0.17407766, -0.52223297,  0.17407766, -0.17407766, -0.52223297],
                             [ 0.52223297,  0.17407766, -0.17407766,  0.52223297,  0.17407766, -0.17407766],
                             [ 0.87038828,  0.52223297,  0.17407766,  0.87038828,  0.52223297, 0.17407766],
                             [ 1.21854359,  0.87038828,  0.52223297,  1.21854359,  0.87038828, 0.52223297]], 
                            dtype=float),
            index   = pd.RangeIndex(start=3, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3',
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3']
        ),
        {1: pd.Series(
                data  = np.array(expected_y_values, dtype=float), 
                index = pd.RangeIndex(start=3, stop=10, step=1),
                name  = f'{level}_step_1'
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        None
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key])
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert isinstance(results[4], type(None))


@pytest.mark.parametrize("level, expected_y_values", 
                         [('l1', ([3., 4., 5., 6., 7., 8.], 
                                  [4., 5., 6., 7., 8., 9.])), 
                          ('l2', ([103., 104., 105., 106., 107., 108.], 
                                  [104., 105., 106., 107., 108., 109.]))])
def test_create_train_X_y_output_when_interspersed_lags_steps_2_and_exog_is_None(level, expected_y_values):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    interspersed lags and steps is 2.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(100, 110), dtype=float)})
    exog = None

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level=level,
                                               lags=[1, 3], steps=2, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[2., 0., 102., 100.],
                             [3., 1., 103., 101.],
                             [4., 2., 104., 102.],
                             [5., 3., 105., 103.],
                             [6., 4., 106., 104.],
                             [7., 5., 107., 105.]], dtype=float),
            index   = pd.RangeIndex(start=4, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_3',
                       'l2_lag_1', 'l2_lag_3']
        ),
        {1: pd.Series(
                data  = np.array(expected_y_values[0], dtype=float), 
                index = pd.RangeIndex(start=3, stop=9, step=1),
                name  = f'{level}_step_1'
            ),
         2: pd.Series(
                data  = np.array(expected_y_values[1], dtype=float), 
                index = pd.RangeIndex(start=4, stop=10, step=1),
                name  = f'{level}_step_2'
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        None
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert isinstance(results[4], type(None))


@pytest.mark.parametrize("level, expected_y_values", 
                         [('l1', ([5., 6., 7., 8.], 
                                  [6., 7., 8., 9.])), 
                          ('l2', ([105., 106., 107., 108.], 
                                  [106., 107., 108., 109.]))])
def test_create_train_X_y_output_when_lags_dict_steps_2_and_exog_is_None(level, expected_y_values):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    different lags and steps is 2.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(100, 110))})
    exog = None

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level=level,
                                               lags={'l1': 3, 'l2': [1, 5]}, 
                                               steps=2, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 104., 100.],
                             [5., 4., 3., 105., 101.],
                             [6., 5., 4., 106., 102.],
                             [7., 6., 5., 107., 103.]], dtype=float),
            index   = pd.RangeIndex(start=6, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3',
                       'l2_lag_1', 'l2_lag_5']
        ),
        {1: pd.Series(
                data  = np.array(expected_y_values[0], dtype=float), 
                index = pd.RangeIndex(start=5, stop=9, step=1),
                name  = f'{level}_step_1'
            ),
         2: pd.Series(
                data  = np.array(expected_y_values[1], dtype=float), 
                index = pd.RangeIndex(start=6, stop=10, step=1),
                name  = f'{level}_step_2'
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        None
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert isinstance(results[4], type(None))


@pytest.mark.parametrize("level, expected_y_values", 
                         [('l1', ([3., 4., 5., 6., 7., 8.], 
                                  [4., 5., 6., 7., 8., 9.])), 
                          ('l2', ([103., 104., 105., 106., 107., 108.], 
                                  [104., 105., 106., 107., 108., 109.]))])
def test_create_train_X_y_output_when_lags_dict_with_None_steps_2_and_exog_is_None(level, expected_y_values):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is a dict with None and steps is 2.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(100, 110))})
    exog = None

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level=level,
                                               lags={'l1': 3, 'l2': None}, 
                                               steps=2, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[2., 1., 0.],
                             [3., 2., 1.],
                             [4., 3., 2.],
                             [5., 4., 3.],
                             [6., 5., 4.],
                             [7., 6., 5.]], dtype=float),
            index   = pd.RangeIndex(start=4, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3']
        ),
        {1: pd.Series(
                data  = np.array(expected_y_values[0], dtype=float), 
                index = pd.RangeIndex(start=3, stop=9, step=1),
                name  = f'{level}_step_1'
            ),
         2: pd.Series(
                data  = np.array(expected_y_values[1], dtype=float), 
                index = pd.RangeIndex(start=4, stop=10, step=1),
                name  = f'{level}_step_2'
            )
        },
        ['l1'] if forecaster.level == 'l1' else ['l1', 'l2'],
        ['l1'],
        None
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert isinstance(results[4], type(None))


def test_create_train_X_y_output_when_y_and_exog_no_pandas_index():
    """
    Test the output of _create_train_X_y when y and exog have no pandas index 
    that doesn't start at 0.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    series.index = np.arange(6, 16)
    exog = pd.Series(np.arange(100, 110), name='exog', 
                     index=np.arange(6, 16), dtype=float)

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                               lags=5, steps=1, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50., 105.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51., 106.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52., 107.],
                             [7., 6., 5., 4., 3., 57., 56., 55., 54., 53., 108.],
                             [8., 7., 6., 5., 4., 58., 57., 56., 55., 54., 109.]], 
                             dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5',
                       'exog_step_1']
        ),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "l1_step_1"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt : f'dtype: {dt}')
def test_create_train_X_y_output_when_lags_5_steps_1_and_exog_is_series_of_float_int(dtype):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 1 and exog is pandas Series of floats or ints.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.Series(np.arange(100, 110), name='exog', dtype=dtype)

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                               lags=5, steps=1, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50., 105.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51., 106.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52., 107.],
                             [7., 6., 5., 4., 3., 57., 56., 55., 54., 53., 108.],
                             [8., 7., 6., 5., 4., 58., 57., 56., 55., 54., 109.]], 
                             dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5',
                       'exog_step_1']
        ).astype({'exog_step_1': dtype}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "l1_step_1"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt : f'dtype: {dt}')
def test_create_train_X_y_output_when_lags_5_steps_2_and_exog_is_series_of_float_int(dtype):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 2 and exog is pandas Series of floats or ints.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.Series(np.arange(100, 110), name='exog', dtype=dtype)

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                               lags=5, steps=2, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)
 
    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50., 105., 106.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51., 106., 107.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52., 107., 108.],
                             [7., 6., 5., 4., 3., 57., 56., 55., 54., 53., 108., 109.]], 
                             dtype=float),
            index = pd.RangeIndex(start=6, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5',
                       'exog_step_1', 'exog_step_2']
        ).astype({'exog_step_1': dtype, 'exog_step_2': dtype}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=9, step=1),
                name  = "l1_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=10, step=1),
                name  = "l1_step_2"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt : f'dtype: {dt}')
def test_create_train_X_y_output_when_lags_5_steps_1_and_exog_is_dataframe_of_float_int(dtype):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 1 and exog is pandas DataFrame of floats or ints.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog   = pd.DataFrame({'exog_1' : np.arange(100, 110, dtype=dtype),
                           'exog_2' : np.arange(1000, 1010, dtype=dtype)})
    
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l2',
                                               lags=5, steps=1, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50., 105., 1005.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51., 106., 1006.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52., 107., 1007.],
                             [7., 6., 5., 4., 3., 57., 56., 55., 54., 53., 108., 1008.],
                             [8., 7., 6., 5., 4., 58., 57., 56., 55., 54., 109., 1009.]], 
                             dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5',
                       'exog_1_step_1', 'exog_2_step_1']
        ).astype({'exog_1_step_1': dtype, 'exog_2_step_1': dtype}),
        {1: pd.Series(
                data  = np.array([55., 56., 57., 58., 59.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "l2_step_1"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog_1', 'exog_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt : f'dtype: {dt}')
def test_create_train_X_y_output_when_lags_5_steps_3_and_exog_is_dataframe_of_float_int(dtype):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 3 and exog is pandas DataFrame of floats or ints.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.DataFrame({'exog_1': np.arange(100, 110, dtype=dtype),
                         'exog_2': np.arange(1000, 1010, dtype=dtype)})
    
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l2',
                                               lags=5, steps=3, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50., 105., 1005., 106., 1006., 107., 1007.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51., 106., 1006., 107., 1007., 108., 1008.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52., 107., 1007., 108., 1008., 109., 1009.]],
                            dtype=float),
            index = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5',
                       'exog_1_step_1', 'exog_2_step_1', 
                       'exog_1_step_2', 'exog_2_step_2', 
                       'exog_1_step_3', 'exog_2_step_3']
        ).astype({'exog_1_step_1': dtype, 'exog_2_step_1': dtype, 
                  'exog_1_step_2': dtype, 'exog_2_step_2': dtype, 
                  'exog_1_step_3': dtype, 'exog_2_step_3': dtype}),
        {1: pd.Series(
                data  = np.array([55., 56., 57.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=8, step=1),
                name  = "l2_step_1"
            ),
         2: pd.Series(
                data  = np.array([56., 57., 58.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=9, step=1),
                name  = "l2_step_2"
            ),
         3: pd.Series(
                data  = np.array([57., 58., 59.], dtype=float), 
                index = pd.RangeIndex(start=7, stop=10, step=1),
                name  = "l2_step_3"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog_1', 'exog_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


@pytest.mark.parametrize("exog_values, dtype", 
                         [([True]    , bool), 
                          (['string'], str)], 
                         ids = lambda dt : f'values, dtype: {dt}')
def test_create_train_X_y_output_when_lags_5_steps_1_and_exog_is_series_of_bool_str(exog_values, dtype):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 1 and exog is pandas Series of bool or str.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.Series(exog_values*10, name='exog', dtype=dtype)

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                               lags=5, steps=1, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52.],
                             [7., 6., 5., 4., 3., 57., 56., 55., 54., 53.],
                             [8., 7., 6., 5., 4., 58., 57., 56., 55., 54.]], 
                             dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5']
        ).assign(exog_step_1=exog_values*5).astype({'exog_step_1': dtype}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "l1_step_1"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


@pytest.mark.parametrize("exog_values, dtype", 
                         [([True]    , bool), 
                          (['string'], str)], 
                         ids = lambda dt : f'values, dtype: {dt}')
def test_create_train_X_y_output_when_lags_5_steps_2_and_exog_is_series_of_bool_str(exog_values, dtype):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 2 and exog is pandas Series of bool or str.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.Series(exog_values*10, name='exog', dtype=dtype)

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                               lags=5, steps=2, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)
 
    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52.],
                             [7., 6., 5., 4., 3., 57., 56., 55., 54., 53.]], 
                             dtype=float),
            index = pd.RangeIndex(start=6, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5']
        ).assign(exog_step_1=exog_values*4,
                 exog_step_2=exog_values*4).astype({'exog_step_1': dtype, 
                                                    'exog_step_2': dtype}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=9, step=1),
                name  = "l1_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=10, step=1),
                name  = "l1_step_2"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key])
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


@pytest.mark.parametrize("v_exog_1   , v_exog_2  , dtype", 
                         [([True]    , [False]   , bool), 
                          (['string'], ['string'], str)], 
                         ids = lambda dt : f'values, dtype: {dt}')
def test_create_train_X_y_output_when_lags_5_steps_1_and_exog_is_dataframe_of_bool_str(v_exog_1, v_exog_2, dtype):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 1 and exog is pandas DataFrame of bool or str.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.DataFrame({'exog_1': v_exog_1*10,
                         'exog_2': v_exog_2*10})
    
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l2',
                                               lags=5, steps=1, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52.],
                             [7., 6., 5., 4., 3., 57., 56., 55., 54., 53.],
                             [8., 7., 6., 5., 4., 58., 57., 56., 55., 54.]], 
                             dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5']
        ).assign(exog_1_step_1=v_exog_1*5,
                 exog_2_step_1=v_exog_2*5).astype({'exog_1_step_1': dtype, 
                                                   'exog_2_step_1': dtype}),
        {1: pd.Series(
                data  = np.array([55., 56., 57., 58., 59.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "l2_step_1"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog_1', 'exog_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


@pytest.mark.parametrize("v_exog_1   , v_exog_2  , dtype", 
                         [([True]    , [False]   , bool), 
                          (['string'], ['string'], str)], 
                         ids = lambda dt : f'values, dtype: {dt}')
def test_create_train_X_y_output_when_lags_5_steps_3_and_exog_is_dataframe_of_bool_str(v_exog_1, v_exog_2, dtype):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 3 and exog is pandas DataFrame of bool or str.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.DataFrame({'exog_1': v_exog_1*10,
                         'exog_2': v_exog_2*10})
    
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l2',
                                               lags=5, steps=3, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52.]],
                            dtype=float),
            index = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5']
        ).assign(exog_1_step_1=v_exog_1*3,
                 exog_2_step_1=v_exog_2*3,
                 exog_1_step_2=v_exog_1*3,
                 exog_2_step_2=v_exog_2*3,
                 exog_1_step_3=v_exog_1*3,
                 exog_2_step_3=v_exog_2*3
        ).astype({'exog_1_step_1': dtype, 'exog_2_step_1': dtype, 
                  'exog_1_step_2': dtype, 'exog_2_step_2': dtype, 
                  'exog_1_step_3': dtype, 'exog_2_step_3': dtype}
        ),
        {1: pd.Series(
                data  = np.array([55., 56., 57.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=8, step=1),
                name  = "l2_step_1"
            ),
         2: pd.Series(
                data  = np.array([56., 57., 58.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=9, step=1),
                name  = "l2_step_2"
            ),
         3: pd.Series(
                data  = np.array([57., 58., 59.], dtype=float), 
                index = pd.RangeIndex(start=7, stop=10, step=1),
                name  = "l2_step_3"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog_1', 'exog_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


def test_create_train_X_y_output_when_lags_5_steps_1_and_exog_is_series_of_category():
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 1 and exog is pandas Series of category.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.Series(range(10), name='exog', dtype='category')

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                               lags=5, steps=1, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52.],
                             [7., 6., 5., 4., 3., 57., 56., 55., 54., 53.],
                             [8., 7., 6., 5., 4., 58., 57., 56., 55., 54.]], 
                             dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5']
        ).assign(exog_step_1=pd.Categorical(range(5, 10), categories=range(10))),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "l1_step_1"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


def test_create_train_X_y_output_when_lags_5_steps_2_and_exog_is_series_of_category():
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 2 and exog is pandas Series of category.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.Series(range(10), name='exog', dtype='category')

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                               lags=5, steps=2, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)
 
    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52.],
                             [7., 6., 5., 4., 3., 57., 56., 55., 54., 53.]], 
                             dtype=float),
            index = pd.RangeIndex(start=6, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5']
        ).assign(exog_step_1=pd.Categorical(range(5, 9), categories=range(10)),
                 exog_step_2=pd.Categorical(range(6, 10), categories=range(10))),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=9, step=1),
                name  = "l1_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=10, step=1),
                name  = "l1_step_2"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


def test_create_train_X_y_output_when_lags_5_steps_1_and_exog_is_dataframe_of_category():
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 1 and exog is pandas DataFrame of category.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.DataFrame({'exog_1': pd.Categorical(range(10)),
                         'exog_2': pd.Categorical(range(100, 110))})
    
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l2',
                                               lags=5, steps=1, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52.],
                             [7., 6., 5., 4., 3., 57., 56., 55., 54., 53.],
                             [8., 7., 6., 5., 4., 58., 57., 56., 55., 54.]], 
                             dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5']
        ).assign(
            exog_1_step_1=pd.Categorical(range(5, 10), categories=range(10)),
            exog_2_step_1=pd.Categorical(range(105, 110), categories=range(100, 110))
        ),
        {1: pd.Series(
                data  = np.array([55., 56., 57., 58., 59.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "l2_step_1"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog_1', 'exog_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


def test_create_train_X_y_output_when_lags_5_steps_3_and_exog_is_dataframe_of_category():
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 3 and exog is pandas DataFrame of category.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.DataFrame({'exog_1': pd.Categorical(range(10)),
                         'exog_2': pd.Categorical(range(100, 110))})
    
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l2',
                                               lags=5, steps=3, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52.]],
                            dtype=float),
            index = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5']
        ).assign(
            exog_1_step_1=pd.Categorical(range(5, 8), categories=range(10)),
            exog_2_step_1=pd.Categorical(range(105, 108), categories=range(100, 110)),
            exog_1_step_2=pd.Categorical(range(6, 9), categories=range(10)),
            exog_2_step_2=pd.Categorical(range(106, 109), categories=range(100, 110)),
            exog_1_step_3=pd.Categorical(range(7, 10), categories=range(10)),
            exog_2_step_3=pd.Categorical(range(107, 110), categories=range(100, 110))
        ),
        {1: pd.Series(
                data  = np.array([55., 56., 57.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=8, step=1),
                name  = "l2_step_1"
            ),
         2: pd.Series(
                data  = np.array([56., 57., 58.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=9, step=1),
                name  = "l2_step_2"
            ),
         3: pd.Series(
                data  = np.array([57., 58., 59.], dtype=float), 
                index = pd.RangeIndex(start=7, stop=10, step=1),
                name  = "l2_step_3"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog_1', 'exog_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


def test_create_train_X_y_output_when_y_is_series_10_and_exog_is_dataframe_of_float_int_category_steps_1():
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 1 and exog is pandas DataFrame of float, int and category.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.DataFrame({'exog_1': pd.Series(np.arange(100, 110), dtype=float),
                         'exog_2': pd.Series(np.arange(1000, 1010), dtype=int),
                         'exog_3': pd.Categorical(range(100, 110))})
    
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                               lags=5, steps=1, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)        
    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50., 105., 1005.],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51., 106., 1006.],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52., 107., 1007.],
                             [7., 6., 5., 4., 3., 57., 56., 55., 54., 53., 108., 1008.],
                             [8., 7., 6., 5., 4., 58., 57., 56., 55., 54., 109., 1009.]], 
                             dtype=float),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5',
                       'exog_1_step_1', 'exog_2_step_1']
        ).astype({'exog_1_step_1': float, 
                  'exog_2_step_1': int}).assign(exog_3_step_1=pd.Categorical(range(105, 110), categories=range(100, 110))
        ),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "l1_step_1"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog_1', 'exog_2', 'exog_3']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


def test_create_train_X_y_output_when_y_is_series_10_and_exog_is_dataframe_of_float_int_category_steps_3():
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 5, steps is 3 and exog is pandas DataFrame of float, int and category.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    exog = pd.DataFrame({'exog_1': pd.Series(np.arange(100, 110), dtype=float),
                         'exog_2': pd.Series(np.arange(1000, 1010), dtype=int),
                         'exog_3': pd.Categorical(range(100, 110))})
    
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1',
                                               lags=5, steps=3, transformer_series=None)
    results = forecaster._create_train_X_y(series=series, exog=exog)        
    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 54., 53., 52., 51., 50., 105, 1005, 105, 106, 1006, 106, 107, 1007, 107],
                             [5., 4., 3., 2., 1., 55., 54., 53., 52., 51., 106, 1006, 106, 107, 1007, 107, 108, 1008, 108],
                             [6., 5., 4., 3., 2., 56., 55., 54., 53., 52., 107, 1007, 107, 108, 1008, 108, 109, 1009, 109]],
                            dtype=float),
            index = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5',
                       'exog_1_step_1', 'exog_2_step_1', 'exog_3_step_1', 
                       'exog_1_step_2', 'exog_2_step_2', 'exog_3_step_2', 
                       'exog_1_step_3', 'exog_2_step_3', 'exog_3_step_3']
        ).astype({'exog_1_step_1': float, 'exog_2_step_1': int,
                  'exog_1_step_2': float, 'exog_2_step_2': int,
                  'exog_1_step_3': float, 'exog_2_step_3': int}
        ).assign(exog_3_step_1=pd.Categorical(range(105, 108), categories=range(100, 110)),
                 exog_3_step_2=pd.Categorical(range(106, 109), categories=range(100, 110)),
                 exog_3_step_3=pd.Categorical(range(107, 110), categories=range(100, 110))
        ),
        {1: pd.Series(
                data  = np.array([5., 6., 7.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=8, step=1),
                name  = "l1_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=9, step=1),
                name  = "l1_step_2"
            ),
         3: pd.Series(
                data  = np.array([7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=7, stop=10, step=1),
                name  = "l1_step_3"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['exog_1', 'exog_2', 'exog_3']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]


@pytest.mark.parametrize("transformer_series", 
                         [StandardScaler(),
                          {'l1': StandardScaler(), 'l2': StandardScaler()}], 
                         ids = lambda tr : f'transformer_series: {tr}')
def test_create_train_X_y_output_when_lags_5_steps_1_and_transformer_series_StandardScaler(transformer_series):
    """
    Test the output of _create_train_X_y when exog is None and transformer_series
    is StandardScaler with steps=1.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(50, 60), dtype=float)})
    
    forecaster = ForecasterAutoregMultiVariate(
                    regressor          = LinearRegression(),
                    lags               = 5,
                    level              = 'l1',
                    steps              = 1,
                    transformer_series = transformer_series
                )
    results = forecaster._create_train_X_y(series=series)

    expected = (
        pd.DataFrame(
            data = np.array([[-0.17407766, -0.52223297, -0.87038828, -1.21854359, -1.5666989 ,
                              -0.17407766, -0.52223297, -0.87038828, -1.21854359, -1.5666989 ],
                             [ 0.17407766, -0.17407766, -0.52223297, -0.87038828, -1.21854359,
                               0.17407766, -0.17407766, -0.52223297, -0.87038828, -1.21854359],
                             [ 0.52223297,  0.17407766, -0.17407766, -0.52223297, -0.87038828,
                               0.52223297,  0.17407766, -0.17407766, -0.52223297, -0.87038828],
                             [ 0.87038828,  0.52223297,  0.17407766, -0.17407766, -0.52223297,
                               0.87038828,  0.52223297,  0.17407766, -0.17407766, -0.52223297],
                             [ 1.21854359,  0.87038828,  0.52223297,  0.17407766, -0.17407766,
                               1.21854359,  0.87038828,  0.52223297,  0.17407766, -0.17407766]]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5']
        ),
        {1: pd.Series(
                data  = np.array([0.17407766, 0.52223297, 0.87038828, 
                                  1.21854359, 1.5666989]), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "l1_step_1"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        None
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert isinstance(results[4], type(None))


@pytest.mark.parametrize("level, expected_y_values", 
                         [('l1', [3., 4., 5., 6., 7., 8., 9.]), 
                          ('l2', [103., 104., 105., 106., 107., 108., 109.])])
def test_create_train_X_y_output_when_lags_3_steps_1_and_exog_is_None_and_transformer_exog_is_not_None(level, expected_y_values):
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 3 and steps is 1 and transformer_exog is not None.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10), dtype=float), 
                           'l2': pd.Series(np.arange(100, 110), dtype=float)})
    exog = None

    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level=level,
                                               lags=3, steps=1, transformer_series=None, 
                                               transformer_exog = StandardScaler())
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[2., 1., 0., 102., 101., 100.],
                             [3., 2., 1., 103., 102., 101.],
                             [4., 3., 2., 104., 103., 102.],
                             [5., 4., 3., 105., 104., 103.],
                             [6., 5., 4., 106., 105., 104.],
                             [7., 6., 5., 107., 106., 105.],
                             [8., 7., 6., 108., 107., 106.]], dtype=float),
            index   = pd.RangeIndex(start=3, stop=10, step=1),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3',
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3']
        ),
        {1: pd.Series(
                data  = np.array(expected_y_values, dtype=float), 
                index = pd.RangeIndex(start=3, stop=10, step=1),
                name  = f'{level}_step_1'
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        None
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert isinstance(results[4], type(None))


@pytest.mark.parametrize("transformer_series", 
                         [StandardScaler(),
                          {'l1': StandardScaler(), 'l2': StandardScaler()}], 
                         ids = lambda tr : f'transformer_series: {tr}')
def test_create_train_X_y_output_when_transformer_series_and_transformer_exog_steps_2(transformer_series):
    """
    Test the output of _create_train_X_y when using transformer_series and 
    transformer_exog with steps=2.
    """
    series = pd.DataFrame({'l1': np.arange(10, dtype=float), 
                           'l2': np.arange(10, dtype=float)},
                           index = pd.date_range("1990-01-01", periods=10, freq='D'))
    exog = pd.DataFrame({
               'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 24.4, 87.2, 47.4, 23.8],
               'col_2': ['a', 'b', 'a', 'b', 'a', 'b', 'a', 'b', 'a', 'b']},
                index = pd.date_range("1990-01-01", periods=10, freq='D'))

    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['col_1']),
                            ('onehot', OneHotEncoder(), ['col_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )

    forecaster = ForecasterAutoregMultiVariate(
                     regressor          = LinearRegression(),
                     lags               = 5,
                     level              = 'l1',
                     steps              = 2,
                     transformer_series = transformer_series,
                     transformer_exog   = transformer_exog
                 )
    results = forecaster._create_train_X_y(series=series, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[-0.17407766, -0.52223297, -0.87038828, -1.21854359, -1.5666989 ,
                              -0.17407766, -0.52223297, -0.87038828, -1.21854359, -1.5666989 ,
                              -0.04719331,  0.        ,   1.        ,-0.81862236,  1.        ,
                               0.        ],
                             [ 0.17407766, -0.17407766, -0.52223297, -0.87038828, -1.21854359,
                               0.17407766, -0.17407766, -0.52223297, -0.87038828, -1.21854359,
                              -0.81862236,  1.        ,  0.        ,  2.03112731,  0.        ,
                               1.        ],
                             [ 0.52223297,  0.17407766, -0.17407766, -0.52223297, -0.87038828,
                               0.52223297,  0.17407766, -0.17407766, -0.52223297, -0.87038828,
                               2.03112731,  0.        ,  1.        ,  0.22507577,  1.        ,
                               0.        ],
                             [ 0.87038828,  0.52223297,  0.17407766, -0.17407766, -0.52223297,
                               0.87038828,  0.52223297,  0.17407766, -0.17407766, -0.52223297,
                               0.22507577,  1.        ,  0.        , -0.84584926,  0.        ,
                               1.        ]]),
            index   = pd.date_range("1990-01-07", periods=4, freq='D'),
            columns = ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_lag_4', 'l1_lag_5', 
                       'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_lag_4', 'l2_lag_5', 
                       'col_1_step_1', 'col_2_a_step_1', 'col_2_b_step_1',
                       'col_1_step_2', 'col_2_a_step_2', 'col_2_b_step_2']
        ),
        {1: pd.Series(
                data  = np.array([0.17407766, 0.52223297, 0.87038828, 1.21854359]), 
                index = pd.date_range("1990-01-06", periods=4, freq='D'),
                name  = "l1_step_1"
            ),
         2: pd.Series(
                data  = np.array([0.52223297, 0.87038828, 1.21854359, 1.5666989]), 
                index = pd.date_range("1990-01-07", periods=4, freq='D'),
                name  = "l1_step_2"
            )
        },
        ['l1', 'l2'],
        ['l1', 'l2'],
        ['col_1', 'col_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]



# Unit test _check_interval
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.utils.utils import check_interval


def test_check_interval_TypeError_when_interval_is_not_a_list():
    """
    Check `TypeError` is raised when `interval` is not a `list`.
    """
    err_msg = re.escape(
                ('`interval` must be a `list`. For example, interval of 95% '
                 'should be as `interval = [2.5, 97.5]`.')
              )
    with pytest.raises(TypeError, match = err_msg):
        check_interval(interval = 'not_a_list')


def test_check_interval_ValueError_when_interval_len_is_not_2():
    """
    Check `ValueError` is raised when `interval` len is not 2.
    """
    err_msg = re.escape(
                ('`interval` must contain exactly 2 values, respectively the '
                 'lower and upper interval bounds. For example, interval of 95% '
                 'should be as `interval = [2.5, 97.5]`.')
              )
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = [2.5, 50.0, 97.5])


def test_check_interval_ValueError_when_interval_lower_bound_less_than_0():
    """
    Check `ValueError` is raised when lower bound is less than 0.
    """
    err_msg = re.escape(
                ('Lower interval bound (-1.0) must be >= 0 and < 100.')
              )
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = [-1.0, 97.5])


@pytest.mark.parametrize("interval", 
                         [[100.0, 97.5], [101.0, 97.5]], 
                         ids = lambda value : f'interval: {value}' )
def test_check_interval_ValueError_when_interval_lower_bound_greater_than_or_equal_to_100(interval):
    """
    Check `ValueError` is raised when lower bound is greater than or equal to 100.
    """
    err_msg = re.escape(
                (f'Lower interval bound ({interval[0]}) must be >= 0 and < 100.')
              )
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = interval)


@pytest.mark.parametrize("interval", 
                         [[2.5, 0.0], [2.5, -1.0]], 
                         ids = lambda value : f'interval: {value}' )
def test_check_interval_ValueError_when_interval_upper_bound_less_than_or_equal_to_0(interval):
    """
    Check `ValueError` is raised when upper bound is less than or equal to 0.
    """
    err_msg = re.escape(
                (f'Upper interval bound ({interval[1]}) must be > 0 and <= 100.')
              )
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = interval)


def test_check_interval_ValueError_when_interval_upper_bound_greater_than_100():
    """
    Check `ValueError` is raised when upper bound is greater than 100.
    """
    err_msg = re.escape(
                ('Upper interval bound (101.0) must be > 0 and <= 100.')
              )
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = [2.5, 101.0])


@pytest.mark.parametrize("interval", 
                         [[2.5, 2.5], [2.5, 2.0]], 
                         ids = lambda value : f'interval: {value}' )
def test_check_interval_ValueError_when_interval_lower_bound_greater_than_or_equal_to_upper_bound(interval):
    """
    Check `ValueError` is raised when lower bound is greater than or equal to
    upper bound.
    """
    err_msg = re.escape(
                (f'Lower interval bound ({interval[0]}) must be less than the '
                 f'upper interval bound ({interval[1]}).')
              )
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = interval)


def test_check_interval_TypeError_when_quantiles_is_not_a_list():
    """
    Check `TypeError` is raised when `quantiles` is not a `list`.
    """
    err_msg = re.escape(
                ("`quantiles` must be a `list`. For example, quantiles 0.05, "
                 "0.5, and 0.95 should be as `quantiles = [0.05, 0.5, 0.95]`.")
            )
    with pytest.raises(TypeError, match = err_msg):
        check_interval(quantiles = 'not_a_list')


@pytest.mark.parametrize("quantiles", 
                         [[-0.01, 0.01, 0.5], [0., 1., 1.1], [-2], [-2, 2]], 
                         ids = lambda value : f'quantiles: {value}' )
def test_check_interval_ValueError_when_elements_in_quantiles_are_out_of_bounds(quantiles):
    """
    Check `ValueError` is raised when any element in `quantiles` is 
    not between 0 and 100.
    """
    err_msg = re.escape(
                ("All elements in `quantiles` must be >= 0 and <= 1.")
            )
    with pytest.raises(ValueError, match = err_msg):
        check_interval(quantiles=quantiles)


def test_check_interval_TypeError_when_alpha_is_not_float():
    """
    Check `TypeError` is raised when `alpha` is not a `float`.
    """
    err_msg = re.escape(
                ('`alpha` must be a `float`. For example, interval of 95% '
                 'should be as `alpha = 0.05`.')
            )
    with pytest.raises(TypeError, match = err_msg):
        check_interval(alpha = 'not_a_float')


@pytest.mark.parametrize("alpha", 
                         [1., 0.], 
                         ids = lambda value : f'alpha: {value}' )
def test_check_interval_ValueError_when_alpha_is_out_of_bounds(alpha):
    """
    Check `ValueError` is raised when alpha is not between 0 and 1.
    """
    err_msg = re.escape(
                f'`alpha` must have a value between 0 and 1. Got {alpha}.'
            )
    with pytest.raises(ValueError, match = err_msg):
        check_interval(alpha=alpha)



# Unit test _create_lags ForecasterAutoregDirect
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.ForecasterAutoregDirect import ForecasterAutoregDirect
from sklearn.linear_model import LinearRegression


def test_check_create_lags_exception_when_n_splits_less_than_0():
    """
    Check exception is raised when n_splits in _create_lags is less than 0.
    """
    forecaster = ForecasterAutoregDirect(LinearRegression(), lags=8, steps=3)
    y = pd.Series(np.arange(10))

    err_msg = re.escape(
                (f"The maximum lag ({forecaster.max_lag}) must be less than the length "
                 f"of the series minus the number of steps ({len(y)-(forecaster.steps-1)}).")
            )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_lags(y=y)

  
def test_create_lags_when_lags_is_3_steps_1_and_y_is_numpy_arange_10():
    """
    Test matrix of lags created properly when lags is 3, steps is 1 and y is
    np.arange(10).
    """
    forecaster = ForecasterAutoregDirect(LinearRegression(), lags=3, steps=1)
    results = forecaster._create_lags(y=np.arange(10))
    expected = (np.array([[2., 1., 0.],
                          [3., 2., 1.],
                          [4., 3., 2.],
                          [5., 4., 3.],
                          [6., 5., 4.],
                          [7., 6., 5.],
                          [8., 7., 6.]]),
                np.array([[3., 4., 5., 6., 7., 8., 9.]])
                )

    assert (results[0] == expected[0]).all()
    assert (results[1] == expected[1]).all()

  
def test_create_lags_when_lags_is_list_interspersed_lags_steps_1_and_y_is_numpy_arange_10():
    """
    Test matrix of lags created properly when lags is a list with interspersed 
    lags, steps is 1 and y is np.arange(10).
    """
    forecaster = ForecasterAutoregDirect(LinearRegression(), lags=[1, 5], steps=1)
    results = forecaster._create_lags(y=np.arange(10))
    expected = (np.array([[4., 0.],
                          [5., 1.],
                          [6., 2.],
                          [7., 3.],
                          [8., 4.]]),
                np.array([[5., 6., 7., 8., 9.]])
                )

    assert (results[0] == expected[0]).all()
    assert (results[1] == expected[1]).all()


def test_create_lags_when_lags_is_3_steps_2_and_y_is_numpy_arange_10():
    """
    Test matrix of lags created properly when lags is 3, steps is 2 and y is
    np.arange(10).
    """
    forecaster = ForecasterAutoregDirect(LinearRegression(), lags=3, steps=2)
    results = forecaster._create_lags(y=np.arange(10))
    expected = (np.array([[2., 1., 0.],
                          [3., 2., 1.],
                          [4., 3., 2.],
                          [5., 4., 3.],
                          [6., 5., 4.],
                          [7., 6., 5.]]),
                np.array([[3., 4., 5., 6., 7., 8.],
                          [4., 5., 6., 7., 8., 9.]])
                )

    assert (results[0] == expected[0]).all()
    assert (results[1] == expected[1]).all()


def test_create_lags_when_lags_is_3_steps_5_and_y_is_numpy_arange_10():
    """
    Test matrix of lags created properly when lags is 3, steps is 5 and y is
    np.arange(10).
    """
    forecaster = ForecasterAutoregDirect(LinearRegression(), lags=3, steps=5)
    results = forecaster._create_lags(y=np.arange(10))
    expected = (np.array([[2., 1., 0.],
                          [3., 2., 1.],
                          [4., 3., 2.]]),
                np.array([[3., 4., 5.],
                          [4., 5., 6.],
                          [5., 6., 7.],
                          [6., 7., 8.],
                          [7., 8., 9.]])
                )

    assert (results[0] == expected[0]).all()
    assert (results[1] == expected[1]).all()


# Unit test set_params ForecasterAutoregMultiSeriesCustom
# ==============================================================================
from skforecast.ForecasterAutoregMultiSeriesCustom import ForecasterAutoregMultiSeriesCustom
from sklearn.linear_model import LinearRegression

def create_predictors(y): # pragma: no cover
    """
    Create first 3 lags of a time series.
    """
    lags = y[-1:-4:-1]

    return lags


def test_set_params():
    """
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                    regressor          = LinearRegression(fit_intercept=True),
                    fun_predictors     = create_predictors,
                    window_size        = 3
                 )
    new_params = {'fit_intercept': False}
    forecaster.set_params(new_params)
    expected = {'copy_X': True,
                'fit_intercept': False,
                'n_jobs': None,
                'normalize': 'deprecated', #For sklearn < 1.2
                'positive': False
               }
    results = forecaster.regressor.get_params()
    results.update({'normalize': 'deprecated'})
    
    assert results == expected


# Unit test predict_bootstrapping ForecasterAutoregDirect
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from skforecast.ForecasterAutoregDirect import ForecasterAutoregDirect
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler

# Fixtures
from .fixtures_ForecasterAutoregDirect import y
from .fixtures_ForecasterAutoregDirect import exog
from .fixtures_ForecasterAutoregDirect import exog_predict


def test_predict_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterAutoregDirect(LinearRegression(), lags=3, steps=5)

    err_msg = re.escape(
                ('This Forecaster instance is not fitted yet. Call `fit` with '
                 'appropriate arguments before using predict.')
              )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict_bootstrapping(steps=5)


def test_predict_bootstrapping_ValueError_when_not_in_sample_residuals_for_some_step():
    """
    Test ValueError is raised when in_sample_residuals=True but there is no
    residuals for some step.
    """
    forecaster = ForecasterAutoregDirect(LinearRegression(), lags=3, steps=2)
    forecaster.fit(y=pd.Series(np.arange(10)))
    forecaster.in_sample_residuals = {2: np.array([1, 2, 3])}

    err_msg = re.escape(
                (f"Not `forecaster.in_sample_residuals` for steps: "
                 f"{set([1, 2]) - set(forecaster.in_sample_residuals.keys())}.")
              )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=None, in_sample_residuals=True)


def test_predict_bootstrapping_ValueError_when_out_sample_residuals_is_None():
    """
    Test ValueError is raised when in_sample_residuals=False and
    forecaster.out_sample_residuals is None.
    """
    forecaster = ForecasterAutoregDirect(LinearRegression(), lags=3, steps=2)
    forecaster.fit(y=pd.Series(np.arange(10)))

    err_msg = re.escape(
                ("`forecaster.out_sample_residuals` is `None`. Use "
                 "`in_sample_residuals=True` or method `set_out_sample_residuals()` "
                 "before `predict_interval()`, `predict_bootstrapping()`, "
                 "`predict_quantiles()` or `predict_dist()`.")
              )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=1, in_sample_residuals=False)


def test_predict_bootstrapping_ValueError_when_not_out_sample_residuals_for_all_steps_predicted():
    """
    Test ValueError is raised when in_sample_residuals=False and
    forecaster.out_sample_residuals is not available for all steps predicted.
    """
    forecaster = ForecasterAutoregDirect(LinearRegression(), lags=3, steps=3)
    forecaster.fit(y=pd.Series(np.arange(15)))
    residuals = {2: np.array([1, 2, 3, 4, 5]), 
                 3: np.array([1, 2, 3, 4, 5])}
    forecaster.out_sample_residuals = residuals

    err_msg = re.escape(
                    (f"Not `forecaster.out_sample_residuals` for steps: "
                     f"{set([1, 2]) - set(forecaster.out_sample_residuals.keys())}. "
                     f"Use method `set_out_sample_residuals()`.")
                )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=[1, 2], in_sample_residuals=False)


def test_predict_bootstrapping_ValueError_when_step_out_sample_residuals_value_is_None():
    """
    Test ValueError is raised when in_sample_residuals=False and
    forecaster.out_sample_residuals has a step with a None.
    """
    forecaster = ForecasterAutoregDirect(LinearRegression(), lags=3, steps=3)
    forecaster.fit(y=pd.Series(np.arange(15)))
    residuals = {1: np.array([1, 2, 3, 4, 5]),
                 2: np.array([1, 2, 3, 4, 5])}
    forecaster.set_out_sample_residuals(residuals = residuals)

    err_msg = re.escape(
                ("forecaster residuals for step 3 are `None`. Check forecaster.out_sample_residuals.")
            )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=3, in_sample_residuals=False)


def test_predict_bootstrapping_ValueError_when_step_out_sample_residuals_value_contains_None():
    """
    Test ValueError is raised when in_sample_residuals=False and
    forecaster.out_sample_residuals has a step with a None value.
    """
    forecaster = ForecasterAutoregDirect(LinearRegression(), lags=3, steps=3)
    forecaster.fit(y=pd.Series(np.arange(15)))
    residuals = {1: np.array([1, 2, 3, 4, 5]),
                 2: np.array([1, 2, 3, 4, 5]), 
                 3: np.array([1, 2, 3, 4, None])}
    forecaster.set_out_sample_residuals(residuals = residuals)

    err_msg = re.escape(
                    ("forecaster residuals for step 3 contains `None` values. Check forecaster.out_sample_residuals.")
                )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=3, in_sample_residuals=False)


@pytest.mark.parametrize("steps", [2, [1, 2], None], 
                         ids=lambda steps: f'steps: {steps}')
def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer(steps):
    """
    Test output of predict_bootstrapping when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterAutoregDirect(
                     regressor        = LinearRegression(),
                     steps            = 2,
                     lags             = 3,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                 )
    forecaster.fit(y=y, exog=exog)
    results = forecaster.predict_bootstrapping(steps=steps, exog=exog_predict, n_boot=4, in_sample_residuals=True)
    expected = pd.DataFrame(
                   data = np.array([[0.73195423, 0.6896871 , 0.20248689, 0.54685553],
                                    [0.58019048, 0.25711887, 0.35256064, 0.82034106]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_False_exog_and_transformer():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterAutoregDirect(
                     regressor        = LinearRegression(),
                     steps            = 2,
                     lags             = 3,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                 )
    forecaster.fit(y=y, exog=exog)
    forecaster.out_sample_residuals = forecaster.in_sample_residuals
    results = forecaster.predict_bootstrapping(steps=2, exog=exog_predict, n_boot=4, in_sample_residuals=False)
    expected = pd.DataFrame(
                   data = np.array([[0.73195423, 0.6896871 , 0.20248689, 0.54685553],
                                    [0.58019048, 0.25711887, 0.35256064, 0.82034106]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_fixed():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals that are fixed.
    """
    forecaster = ForecasterAutoregDirect(
                     regressor = LinearRegression(),
                     steps     = 2,
                     lags      = 3
                 )
    forecaster.fit(y=y, exog=exog)
    forecaster.in_sample_residuals = {1: pd.Series([1, 1, 1, 1, 1, 1, 1]),
                                      2: pd.Series([5, 5, 5, 5, 5, 5, 5])}
    results = forecaster.predict_bootstrapping(steps=2, exog=exog_predict, n_boot=4, in_sample_residuals=True)
    expected = pd.DataFrame(
                   data = np.array([[1.67523588, 1.67523588, 1.67523588, 1.67523588],
                                    [5.38024988, 5.38024988, 5.38024988, 5.38024988]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    
    pd.testing.assert_frame_equal(expected, results)        


# Unit test _evaluate_grid_hyperparameters
# ==============================================================================
import re
import os
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import Ridge
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_absolute_percentage_error
from skforecast.ForecasterAutoreg import ForecasterAutoreg
from skforecast.ForecasterAutoregCustom import ForecasterAutoregCustom
from skforecast.model_selection.model_selection import _evaluate_grid_hyperparameters

from tqdm import tqdm
from functools import partialmethod
tqdm.__init__ = partialmethod(tqdm.__init__, disable=True) # hide progress bar

# Fixtures
from .fixtures_model_selection import y

def create_predictors(y): # pragma: no cover
    """
    Create first 4 lags of a time series, used in ForecasterAutoregCustom.
    """

    lags = y[-1:-5:-1]

    return lags


def test_ValueError_evaluate_grid_hyperparameters_when_return_best_and_len_y_exog_different():
    """
    Test ValueError is raised in _evaluate_grid_hyperparameters when return_best 
    and length of `y` and `exog` do not match.
    """
    forecaster = ForecasterAutoreg(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    exog = y[:30]

    err_msg = re.escape(
            (f"`exog` must have same number of samples as `y`. "
             f"length `exog`: ({len(exog)}), length `y`: ({len(y)})")
        )
    with pytest.raises(ValueError, match = err_msg):
        _evaluate_grid_hyperparameters(
            forecaster         = forecaster,
            y                  = y,
            exog               = exog,
            lags_grid          = [2, 4],
            param_grid         = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}],
            steps              = 3,
            refit              = False,
            metric             = 'mean_absolute_error',
            initial_train_size = len(y[:-12]),
            fixed_train_size   = False,
            return_best        = True,
            verbose            = False
        )


def test_ValueError_evaluate_grid_hyperparameters_metric_list_duplicate_names():
    """
    Test ValueError is raised in _evaluate_grid_hyperparameters when a `list` of 
    metrics is used with duplicate names.
    """
    forecaster = ForecasterAutoreg(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )

    err_msg = re.escape('When `metric` is a `list`, each metric name must be unique.')
    with pytest.raises(ValueError, match = err_msg):
        _evaluate_grid_hyperparameters(
            forecaster         = forecaster,
            y                  = y,
            lags_grid          = [2, 4],
            param_grid         = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}],
            steps              = 3,
            refit              = False,
            metric             = ['mean_absolute_error', mean_absolute_error],
            initial_train_size = len(y[:-12]),
            fixed_train_size   = False,
            return_best        = False,
            verbose            = False
        )


def test_output_evaluate_grid_hyperparameters_ForecasterAutoreg_with_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterAutoreg with mocked
    (mocked done in Skforecast v0.4.3).
    """
    forecaster = ForecasterAutoreg(
                     regressor = Ridge(random_state=123),
                     lags      = 2 
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    idx = len(lags_grid)*len(param_grid)

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  lags_grid          = lags_grid,
                  param_grid         = param_grid,
                  steps              = steps,
                  refit              = False,
                  metric             = 'mean_squared_error',
                  initial_train_size = len(y_train),
                  fixed_train_size   = False,
                  return_best        = False,
                  verbose            = False
              )
    
    expected_results = pd.DataFrame({
            'lags'       : [[1, 2], [1, 2], [1, 2], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]],
            'lags_label' : [[1, 2], [1, 2], [1, 2], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]],
            'params'     : [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}, 
                            {'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}],
            'mean_squared_error':np.array([0.06464646, 0.06502362, 0.06745534, 
                                           0.06779272, 0.06802481, 0.06948609]),                                                               
            'alpha'      : np.array([0.01, 0.1 , 1.  , 0.01, 0.1 , 1.  ])
                                     },
            index = pd.RangeIndex(start=0, stop=idx, step=1)
    )

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_evaluate_grid_hyperparameters_ForecasterAutoreg_with_diferentiation_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterAutoreg with mocked
    (mocked done in Skforecast v0.10.0) when differentiation is used.
    """
    forecaster = ForecasterAutoreg(
                     regressor       = Ridge(random_state=123),
                     lags            = 2, # Placeholder, the value will be overwritten
                     differentiation = 1
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    idx = len(lags_grid)*len(param_grid)

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  lags_grid          = lags_grid,
                  param_grid         = param_grid,
                  steps              = steps,
                  refit              = False,
                  metric             = 'mean_squared_error',
                  initial_train_size = len(y_train),
                  fixed_train_size   = False,
                  return_best        = False,
                  verbose            = False
              ).reset_index(drop=True)
    
    expected_results = pd.DataFrame({
        'lags'       : [[1, 2, 3, 4], [1, 2], [1, 2, 3, 4], [1, 2], [1, 2], [1, 2, 3, 4]],
        'lags_label' : [[1, 2, 3, 4], [1, 2], [1, 2, 3, 4], [1, 2], [1, 2], [1, 2, 3, 4]],
        'params'     : [{'alpha': 1}, {'alpha': 1}, {'alpha': 0.1}, 
                        {'alpha': 0.1}, {'alpha': 0.01}, {'alpha': 0.01}],
        'mean_squared_error': np.array([0.09168123, 0.09300068, 0.09930084, 
                                        0.09960109, 0.10102995, 0.1012931]),                                                               
        'alpha'      : np.array([1., 1., 0.1, 0.1, 0.01 , 0.01])
        },
        index = pd.RangeIndex(start=0, stop=idx, step=1)
    )

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_evaluate_grid_hyperparameters_ForecasterAutoregCustom_with_diferentiation_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterAutoregCustom with mocked
    (mocked done in Skforecast v0.10.0) when differentiation is used.
    """
    forecaster = ForecasterAutoregCustom(
                     regressor       = Ridge(random_state=123),
                     fun_predictors  = create_predictors,
                     window_size     = 4,
                     differentiation = 1
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    idx = len(param_grid)

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  param_grid         = param_grid,
                  steps              = steps,
                  refit              = False,
                  metric             = 'mean_squared_error',
                  initial_train_size = len(y_train),
                  fixed_train_size   = False,
                  return_best        = False,
                  verbose            = False
              ).reset_index(drop=True)
    
    expected_results = pd.DataFrame({
        'lags'       : ['custom predictors', 'custom predictors', 'custom predictors'],
        'lags_label' : ['custom predictors', 'custom predictors', 'custom predictors'],
        'params'     : [{'alpha': 1}, {'alpha': 0.1}, {'alpha': 0.01}],
        'mean_squared_error': np.array([0.09168123, 0.09930084, 0.1012931]),                                                               
        'alpha'      : np.array([1., 0.1, 0.01])
        },
        index = pd.RangeIndex(start=0, stop=idx, step=1)
    )

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_evaluate_grid_hyperparameters_ForecasterAutoreg_lags_grid_dict_with_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterAutoreg when 
    `lags_grid` is a dict with mocked (mocked done in Skforecast v0.4.3).
    """
    forecaster = ForecasterAutoreg(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = {'lags_1': 2, 'lags_2': 4}
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    idx = len(lags_grid)*len(param_grid)

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  lags_grid          = lags_grid,
                  param_grid         = param_grid,
                  steps              = steps,
                  refit              = False,
                  metric             = 'mean_squared_error',
                  initial_train_size = len(y_train),
                  fixed_train_size   = False,
                  return_best        = False,
                  verbose            = False
              )
    
    expected_results = pd.DataFrame({
        'lags'       : [[1, 2], [1, 2], [1, 2], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]],
        'lags_label' : ['lags_1', 'lags_1', 'lags_1', 'lags_2', 'lags_2', 'lags_2'],
        'params'     : [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}, {'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}],
        'mean_squared_error': np.array([0.06464646, 0.06502362, 0.06745534, 0.06779272, 0.06802481, 0.06948609]),                                                               
        'alpha'      : np.array([0.01, 0.1 , 1.  , 0.01, 0.1 , 1.  ])
        },
        index = pd.RangeIndex(start=0, stop=idx, step=1)
    )

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_evaluate_grid_hyperparameters_ForecasterAutoreg_lags_grid_is_None_with_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterAutoreg when 
    `lags_grid` is None with mocked (mocked done in Skforecast v0.4.3), 
    should use forecaster.lags as lags_grid.
    """
    forecaster = ForecasterAutoreg(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = None
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  lags_grid          = lags_grid,
                  param_grid         = param_grid,
                  steps              = steps,
                  refit              = False,
                  metric             = 'mean_squared_error',
                  initial_train_size = len(y_train),
                  fixed_train_size   = False,
                  return_best        = False,
                  verbose            = False
              )
    
    expected_results = pd.DataFrame({
        'lags'       : [[1, 2], [1, 2], [1, 2]],
        'lags_label' : [[1, 2], [1, 2], [1, 2]],
        'params'     : [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}],
        'mean_squared_error': np.array([0.06464646, 0.06502362, 0.06745534]),                                                               
        'alpha'      : np.array([0.01, 0.1 , 1.])
        },
        index = pd.RangeIndex(start=0, stop=3, step=1)
    )

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_evaluate_grid_hyperparameters_ForecasterAutoreg_metric_list_with_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterAutoreg with mocked
    and multiple metrics (mocked done in Skforecast v0.4.3).
    """
    forecaster = ForecasterAutoreg(
                     regressor = Ridge(random_state=123),
                     lags      = 2 # Placeholder, the value will be overwritten
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    idx = len(lags_grid)*len(param_grid)

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  lags_grid          = lags_grid,
                  param_grid         = param_grid,
                  steps              = steps,
                  refit              = False,
                  metric             = ['mean_squared_error', mean_absolute_error],
                  initial_train_size = len(y_train),
                  fixed_train_size   = False,
                  return_best        = False,
                  verbose            = False
              )
    
    expected_results = pd.DataFrame({
        'lags'       : [[1, 2], [1, 2], [1, 2], [1, 2, 3, 4], 
                        [1, 2, 3, 4], [1, 2, 3, 4]],
        'lags_label' : [[1, 2], [1, 2], [1, 2], [1, 2, 3, 4], 
                        [1, 2, 3, 4], [1, 2, 3, 4]],
        'params'     : [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}, 
                        {'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}],
        'mean_squared_error': np.array([0.06464646, 0.06502362, 0.06745534, 
                                        0.06779272, 0.06802481, 0.06948609]),   
        'mean_absolute_error': np.array([0.20278812, 0.20314819, 0.20519952, 
                                         0.20601567, 0.206323, 0.20747017]),                                                          
        'alpha'      : np.array([0.01, 0.1 , 1.  , 0.01, 0.1 , 1.  ])},
        index = pd.RangeIndex(start=0, stop=idx, step=1)
    )

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_evaluate_grid_hyperparameters_ForecasterAutoregCustom_with_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterAutoregCustom with mocked
    (mocked done in Skforecast v0.4.3).
    """
    forecaster = ForecasterAutoregCustom(
                     regressor      = Ridge(random_state=123),
                     fun_predictors = create_predictors,
                     window_size    = 4
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    idx = len(param_grid)

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  param_grid         = param_grid,
                  steps              = steps,
                  refit              = False,
                  metric             = 'mean_squared_error',
                  initial_train_size = len(y_train),
                  fixed_train_size   = False,
                  return_best        = False,
                  verbose            = False
              )
    
    expected_results = pd.DataFrame({
        'lags'       : ['custom predictors', 'custom predictors', 'custom predictors'],
        'lags_label' : ['custom predictors', 'custom predictors', 'custom predictors'],
        'params'     : [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}],
        'mean_squared_error': np.array([0.06779272, 0.06802481, 0.06948609]),                                                               
        'alpha'      : np.array([0.01, 0.1 , 1.])
        },
        index = pd.RangeIndex(start=0, stop=idx, step=1)
     )
    
    pd.testing.assert_frame_equal(results, expected_results)


@pytest.mark.parametrize("lags_grid", 
                         [[2, 4], {'lags_1': 2, 'lags_2': 4}], 
                         ids=lambda lg: f'lags_grid: {lg}')
def test_evaluate_grid_hyperparameters_when_return_best_ForecasterAutoreg(lags_grid):
    """
    Test forecaster is refitted when return_best=True in 
    _evaluate_grid_hyperparameters with ForecasterAutoreg.
    """
    forecaster = ForecasterAutoreg(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]

    _evaluate_grid_hyperparameters(
        forecaster         = forecaster,
        y                  = y,
        lags_grid          = lags_grid,
        param_grid         = param_grid,
        steps              = steps,
        refit              = False,
        metric             = 'mean_squared_error',
        initial_train_size = len(y_train),
        fixed_train_size   = False,
        return_best        = True,
        verbose            = False
    )
    
    expected_lags = np.array([1, 2])
    expected_alpha = 0.01
    
    assert (expected_lags == forecaster.lags).all()
    assert expected_alpha == forecaster.regressor.alpha


def test_evaluate_grid_hyperparameters_when_return_best_ForecasterAutoregCustom():
    """
    Test forecaster is refitted when return_best=True in 
    _evaluate_grid_hyperparameters with ForecasterAutoregCustom.
    """
    forecaster = ForecasterAutoregCustom(
                     regressor      = Ridge(random_state=123),
                     fun_predictors = create_predictors,
                     window_size    = 4
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    lags_grid = None # Ignored

    _evaluate_grid_hyperparameters(
        forecaster         = forecaster,
        y                  = y,
        lags_grid          = lags_grid,
        param_grid         = param_grid,
        steps              = steps,
        refit              = False,
        metric             = 'mean_squared_error',
        initial_train_size = len(y_train),
        fixed_train_size   = False,
        return_best        = True,
        verbose            = False
    )
    
    expected_alpha = 0.01
    
    assert expected_alpha == forecaster.regressor.alpha


@pytest.mark.parametrize("lags_grid", 
                         [[2, 4], {'lags_1': 2, 'lags_2': 4}], 
                         ids=lambda lg: f'lags_grid: {lg}')
def test_evaluate_grid_hyperparameters_when_return_best_and_list_metrics(lags_grid):
    """
    Test forecaster is refitted when return_best=True in _evaluate_grid_hyperparameters
    and multiple metrics.
    """
    forecaster = ForecasterAutoreg(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]

    _evaluate_grid_hyperparameters(
        forecaster         = forecaster,
        y                  = y,
        lags_grid          = lags_grid,
        param_grid         = param_grid,
        steps              = steps,
        refit              = False,
        metric             = [mean_absolute_percentage_error, 'mean_squared_error'],
        initial_train_size = len(y_train),
        fixed_train_size   = False,
        return_best        = True,
        verbose            = False,
        show_progress      = False
    )
    
    expected_lags = np.array([1, 2])
    expected_alpha = 1.
    
    assert (expected_lags == forecaster.lags).all()
    assert expected_alpha == forecaster.regressor.alpha


def test_evaluate_grid_hyperparameters_output_file_when_single_metric():
    """
    Test output file is created when output_file is passed to
    _evaluate_grid_hyperparameters and single metric.
    """
    forecaster = ForecasterAutoreg(
                     regressor = Ridge(random_state=123),
                     lags      = 2 
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    output_file = 'test_evaluate_grid_hyperparameters_output_file.txt'

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  lags_grid          = lags_grid,
                  param_grid         = param_grid,
                  steps              = steps,
                  refit              = False,
                  metric             = 'mean_squared_error',
                  initial_train_size = len(y_train),
                  fixed_train_size   = False,
                  return_best        = False,
                  verbose            = False,
                  show_progress      = False,
                  output_file        = output_file
              )
    results  = results.astype({'lags': str, 'lags_label': str, 'params': str})

    assert os.path.isfile(output_file)
    output_file_content = pd.read_csv(output_file, sep='\t', low_memory=False)
    output_file_content = output_file_content.sort_values(by='mean_squared_error')
    output_file_content = output_file_content.astype({'lags': str, 'lags_label': str, 'params': str})
    pd.testing.assert_frame_equal(results, output_file_content)
    os.remove(output_file)


def test_evaluate_grid_hyperparameters_output_file_when_single_metric_as_list():
    """ 
    Test output file is created when output_file is passed to 
    _evaluate_grid_hyperparameters and single metric as list.
    """
    forecaster = ForecasterAutoreg(
                     regressor = Ridge(random_state=123),
                     lags      = 2 
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    output_file = 'test_evaluate_grid_hyperparameters_output_file.txt'

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  lags_grid          = lags_grid,
                  param_grid         = param_grid,
                  steps              = steps,
                  refit              = False,
                  metric             = ['mean_squared_error'],
                  initial_train_size = len(y_train),
                  fixed_train_size   = False,
                  return_best        = False,
                  verbose            = False,
                  show_progress      = False,
                  output_file        = output_file
              )
    results  = results.astype({'lags': str, 'lags_label': str, 'params': str})

    assert os.path.isfile(output_file)
    output_file_content = pd.read_csv(output_file, sep='\t', low_memory=False)
    output_file_content = output_file_content.sort_values(by='mean_squared_error')
    output_file_content = output_file_content.astype({'lags': str, 'lags_label': str, 'params': str})
    pd.testing.assert_frame_equal(results, output_file_content)
    os.remove(output_file)


def test_evaluate_grid_hyperparameters_output_file_when_2_metrics_as_list():
    """
    Test output file is created when output_file is passed to 
    _evaluate_grid_hyperparameters and 2 metrics as list.
    """
    forecaster = ForecasterAutoreg(
                     regressor = Ridge(random_state=123),
                     lags      = 2 
                 )

    steps = 3
    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    output_file = 'test_evaluate_grid_hyperparameters_output_file.txt'

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  lags_grid          = lags_grid,
                  param_grid         = param_grid,
                  steps              = steps,
                  refit              = False,
                  metric             = ['mean_squared_error', 'mean_absolute_error'],
                  initial_train_size = len(y_train),
                  fixed_train_size   = False,
                  return_best        = False,
                  verbose            = False,
                  show_progress      = False,
                  output_file        = output_file
              )
    results  = results.astype({'lags': str, 'lags_label': str, 'params': str})

    assert os.path.isfile(output_file)
    output_file_content = pd.read_csv(output_file, sep='\t', low_memory=False)
    output_file_content = output_file_content.sort_values(by='mean_squared_error')
    output_file_content = output_file_content.astype({'lags': str, 'lags_label': str, 'params': str})
    pd.testing.assert_frame_equal(results, output_file_content)
    os.remove(output_file)


# Unit test _binning_in_sample_residuals ForecasterAutoreg
# ==============================================================================
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from skforecast.ForecasterAutoreg import ForecasterAutoreg


def test_binning_in_sample_residuals_output():
    """
    Test that _binning_in_sample_residuals returns the expected output.
    """

    forecaster = ForecasterAutoreg(
        regressor=object(),
        lags = 5,
        binner_kwargs={'n_bins':3}
    )

    rng = np.random.default_rng(12345)
    y_pred = pd.Series(rng.normal(100, 15, 20))
    y_true = pd.Series(rng.normal(100, 10, 20))

    forecaster._binning_in_sample_residuals(
        y_pred=y_pred,
        y_true=y_true,
    )

    expected_1 = np.array([
                    34.58035615,  22.08911948,  15.6081091 ,   7.0808798 ,
                    55.47454021,   1.80092458,  12.2624188 , -12.32822882,
                    -0.451743  ,  11.83152763,  -7.11862253,   6.32581108,
                    -20.92461257, -21.95291202, -10.55026794, -27.43753138,
                    -6.24112163, -25.62685698,  -4.31327121, -18.40910724
                ])

    expected_2 = {
        0: np.array([
                34.58035615, 22.08911948, 15.6081091 ,  7.0808798 , 55.47454021,
                1.80092458, 12.2624188
            ]),
        1: np.array([
            -12.32822882,  -0.451743  ,  11.83152763,  -7.11862253,
            6.32581108, -20.92461257
            ]),
        2: np.array([
            -21.95291202, -10.55026794, -27.43753138,  -6.24112163,
            -25.62685698,  -4.31327121, -18.40910724
            ])
    }

    expected_3 = {
        0: (70.70705405481715, 90.25638761254116),
        1: (90.25638761254116, 109.36821559391004),
        2: (109.36821559391004, 135.2111448156828)
    }

    np.testing.assert_almost_equal(forecaster.in_sample_residuals, expected_1)

    for k in expected_2.keys():
        np.testing.assert_almost_equal(forecaster.in_sample_residuals_by_bin[k], expected_2[k])

    assert forecaster.binner_intervals == expected_3


def test_binning_in_sample_residuals_stores_maximum_200_residuals_per_bin():
    """
    Test that _binning_in_sample_residuals stores a maximum of 200 residuals per bin.
    """

    y = pd.Series(
            data = np.random.normal(loc=10, scale=1, size=1000),
            index = pd.date_range(start='01-01-2000', periods=1000, freq='H')

        )
    forecaster = ForecasterAutoreg(
                    regressor=LinearRegression(),
                    lags = 5,
                    binner_kwargs={'n_bins':2}
                )
    forecaster.fit(y)

    for v in forecaster.in_sample_residuals_by_bin.values():
        assert len(v) == 200
        assert len(v) > 0

    np.testing.assert_array_almost_equal(
        forecaster.in_sample_residuals,
        np.concatenate(list(forecaster.in_sample_residuals_by_bin.values()))
    )


# Unit test create_train_X_y ForecasterAutoregMultiSeries
# ==============================================================================
import numpy as np
import pandas as pd
from skforecast.ForecasterAutoregMultiSeries import ForecasterAutoregMultiSeries
from sklearn.linear_model import LinearRegression


def test_create_train_X_y_output_when_series_and_exog_is_None():
    """
    Test the output of create_train_X_y when series has 2 columns and 
    exog is None.
    """
    series = pd.DataFrame({'1': pd.Series(np.arange(7, dtype=float)), 
                           '2': pd.Series(np.arange(7, dtype=float))})
    forecaster = ForecasterAutoregMultiSeries(
        regressor = LinearRegression(),
        lags      = 3,
        encoding  = "onehot"
    )

    results = forecaster.create_train_X_y(series=series)
    expected = (
        pd.DataFrame(
            data = np.array([[-0.5, -1. , -1.5, 1., 0.],
                             [ 0. , -0.5, -1. , 1., 0.],
                             [ 0.5,  0. , -0.5, 1., 0.],
                             [ 1. ,  0.5,  0. , 1., 0.],
                             [-0.5, -1. , -1.5, 0., 1.],
                             [ 0. , -0.5, -1. , 0., 1.],
                             [ 0.5,  0. , -0.5, 0., 1.],
                             [ 1. ,  0.5,  0. , 0., 1.]]),
            index   = pd.Index([3, 4, 5, 6, 3, 4, 5, 6]),
            columns = ['lag_1', 'lag_2', 'lag_3', '1', '2']
        ).astype({'1': int, '2': int}),
        pd.Series(
            data  = np.array([0., 0.5, 1., 1.5, 0., 0.5, 1., 1.5]),
            index = pd.Index([3, 4, 5, 6, 3, 4, 5, 6]),
            name  = 'y',
            dtype = float
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_series_equal(results[1], expected[1])


from .model_selection import (
    backtesting_forecaster,
    grid_search_forecaster,
    random_search_forecaster,
    bayesian_search_forecaster,
    _create_backtesting_folds,
    select_features
)